"""stream_aggregate example.

docker run --rm -p 9092:9092 emgeee/kafka_emit_measurements:latest
"""

import json
import pprint as pp
import signal
import sys

from denormalized import Context
from denormalized.datafusion import col
from denormalized.datafusion import functions as f


def signal_handler(_sig, _frame):
    sys.exit(0)


signal.signal(signal.SIGINT, signal_handler)

bootstrap_server = "localhost:9092"
timestamp_column = "occurred_at_ms"

sample_event = {
    "occurred_at_ms": 100,
    "sensor_name": "foo",
    "reading": 0.0,
}


def print_batch(rb):
    pp.pprint(rb.to_pydict())


ctx = Context()
temperature_ds = ctx.from_topic(
    "temperature", json.dumps(sample_event), bootstrap_server, timestamp_column
)

humidity_ds = (
    ctx.from_topic(
        "humidity",
        json.dumps(sample_event),
        bootstrap_server,
        timestamp_column,
    )
    .with_column("humidity_sensor", col("sensor_name"))
    .drop_columns(["sensor_name"])
    .window(
        [col("humidity_sensor")],
        [
            f.count(col("reading")).alias("avg_humidity"),
        ],
        4000,
        None,
    )
    .with_column("humidity_window_start_time", col("window_start_time"))
    .with_column("humidity_window_end_time", col("window_end_time"))
    .drop_columns(["window_start_time", "window_end_time"])
)

joined_ds = (
    temperature_ds.window(
        [col("sensor_name")],
        [
            f.avg(col("reading")).alias("avg_temperature"),
        ],
        4000,
        None,
    )
    .join(
        humidity_ds,
        "inner",
        ["sensor_name", "window_start_time"],
        ["humidity_sensor", "humidity_window_start_time"],
    )
    .sink(print_batch)
)



"""stream_aggregate example."""

import json
import signal
import sys
from collections import Counter
from typing import List
import pyarrow as pa

from denormalized import Context
from denormalized.datafusion import Accumulator, col
from denormalized.datafusion import functions as f
from denormalized.datafusion import udaf


def signal_handler(sig, frame):
    sys.exit(0)


signal.signal(signal.SIGINT, signal_handler)

bootstrap_server = "localhost:9092"

sample_event = {
    "occurred_at_ms": 100,
    "sensor_name": "foo",
    "reading": 0.0,
}

class TotalValuesRead(Accumulator):
    # Define the state type as a struct containing a map
    acc_state_type = pa.struct([("counts", pa.map_(pa.string(), pa.int64()))])

    def __init__(self):
        self.counts = Counter()

    def update(self, values: pa.Array) -> None:
        # Update counter with new values
        if values is not None:
            self.counts.update(values.to_pylist())

    def merge(self, states: pa.Array) -> None:
        # Merge multiple states into this accumulator
        if states is None or len(states) == 0:
            return
        for state in states:
            if state is not None:
                counts_map = state.to_pylist()[0] # will always be one element struct
                for k, v in counts_map["counts"]:
                    self.counts[k] += v

    def state(self) -> List[pa.Scalar]:
        # Convert current state to Arrow array format
        result = {"counts": dict(self.counts.items())}
        return [pa.scalar(result, type=pa.struct([("counts", pa.map_(pa.string(), pa.int64()))]))]

    def evaluate(self) -> pa.Scalar:
        return self.state()[0]


input_type = [pa.string()]
return_type = TotalValuesRead.acc_state_type
state_type = [TotalValuesRead.acc_state_type]
sample_udaf = udaf(TotalValuesRead, input_type, return_type, state_type, "stable")


def print_batch(rb: pa.RecordBatch):
    if not len(rb):
        return
    print(rb)

ctx = Context()
ds = ctx.from_topic("temperature", json.dumps(sample_event), bootstrap_server, "occurred_at_ms")

ds = ds.window(
    [],
    [
        sample_udaf(col("sensor_name")).alias("count"),
    ],
    2000,
    None,
).sink(print_batch)



from denormalized._d_internal import PyContext

from .data_stream import DataStream


class Context:
    """Context."""

    def __init__(self) -> None:
        """__init__."""
        self.ctx = PyContext()

    def __repr__(self):
        return self.ctx.__repr__()

    def __str__(self):
        return self.ctx.__str__()

    def from_topic(
        self,
        topic: str,
        sample_json: str,
        bootstrap_servers: str,
        timestamp_column: str,
        group_id: str = "default_group",
    ) -> DataStream:
        """Create a new context from a topic."""
        py_ds = self.ctx.from_topic(
            topic,
            sample_json,
            bootstrap_servers,
            timestamp_column,
            group_id,
        )
        ds = DataStream(py_ds)

        return ds



# feast_data_stream.pyi
from typing import List, Optional, Type, Union

from denormalized._internal import PyDataStream
from denormalized.datafusion import Expr
from feast import FeatureStore, Field

from .data_stream import DataStream

class FeastDataStreamMeta(type):
    def __new__(cls, name: str, bases: tuple, attrs: dict) -> Type[FeastDataStream]: ...

class FeastDataStream(DataStream, metaclass=FeastDataStreamMeta):
    def __init__(self, stream: Union[PyDataStream, DataStream]) -> None: ...
    def select(self, expr_list: List[Expr]) -> FeastDataStream: ...
    def filter(self, predicate: Expr) -> FeastDataStream: ...
    def with_column(self, name: str, predicate: Expr) -> FeastDataStream: ...
    def join_on(
        self, right: DataStream, join_type: str, on_exprs: List[Expr]
    ) -> FeastDataStream: ...
    def join(
        self,
        right: DataStream,
        join_type: str,
        left_cols: List[str],
        right_cols: List[str],
        filter: Optional[Expr] = None,
    ) -> FeastDataStream: ...
    def window(
        self,
        group_exprs: List[Expr],
        aggr_exprs: List[Expr],
        window_length_millis: int,
        slide_millis: Optional[int] = None,
    ) -> FeastDataStream: ...
    def print_schema(self) -> FeastDataStream: ...
    def print_plan(self) -> FeastDataStream: ...
    def print_physical_plan(self) -> FeastDataStream: ...
    def get_feast_schema(self) -> List[Field]: ...
    def write_feast_feature(
        self, feature_store: FeatureStore, source_name: str
    ) -> None: ...
